/****************************************************************************/
/*                                                                          */
/*                   Copyright 2006 University of Glasgow                   */
/*                                                                          */
/*                           All rights reserved.                           */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/* FILE          : Checker.java                                             */
/* DESCRIPTION   : Typechecker for Functional Language with Sessions        */
/*                                                                          */
/* AUTHOR        : S.J.Gay                                                  */
/* DATE          : 22nd December 2005                                       */
/*                                                                          */
/* LAST MODIFIED : S.J.Gay                                                  */
/* DATE          : 4th January 2006                                         */
/****************************************************************************/

package fls.checker;

import fls.node.*;
import fls.analysis.*;
import fls.error.*;
import fls.types.*;
import fls.env.*;
import fls.typenv.*;

import java.util.*;
import java.io.*;

public class Checker extends DepthFirstAdapter {

    ErrorTable errorTable = new ErrorTable();
    HashSet<String> typeNames;
    TypeEnv tenv = new TypeEnv();
    Env env = new Env();
    Hashtable<Node,Integer> position = new Hashtable<Node,Integer>();

    public void setTypeNames(HashSet<String> tn) {
	typeNames = tn;
    }

    public ErrorTable getErrorTable() {
	return errorTable;
    }

    public Env getEnv() {
	return env;
    }

    public void outStart(Start node) {
	setOut(node, getOut(node.getPProgram()));
	position.put(node,position.get(node.getPProgram()));
    }

    public void outAProgram(AProgram node) {
	setOut(node, getOut(node.getExpression()));
	position.put(node,node.getTypes().getLine());
    }

    public void outATermExpression(ATermExpression node)
    {
        setOut(node, getOut(node.getTerm()));
	position.put(node,position.get(node.getTerm()));
    }

    public void caseACondExpression(ACondExpression node)
    {
	node.getCond().apply(this);
	Env envCopy = env.copy();
	node.getThenBranch().apply(this);
	Env envAfterThen = env.copy();
	env = envCopy;
	node.getElseBranch().apply(this);
	if (!env.equal(envAfterThen)) {
	    errorTable.add(node.getIf().getLine(),new CondEnvError());
	};

	Type condType = (Type)getOut(node.getCond());
	Type thenType = (Type)getOut(node.getThenBranch());
	Type elseType = (Type)getOut(node.getElseBranch());

	if (condType != null) {
	    if (!(condType instanceof BoolType)) {
		errorTable.add(node.getIf().getLine(),
			       new IfCondError(condType.name()));
	    }
	};
	if ((thenType != null) && (elseType != null)) {
	    if (!(thenType.equal(elseType))) {
		errorTable.add(node.getIf().getLine(),
		    new IfMismatchError(thenType.name(),elseType.name()));
	    }
	    else {
		if (condType instanceof BoolType) {
		    setOut(node,thenType);
		}
	    }
	};
	position.put(node,node.getIf().getLine());
    }

    public void caseAFnExpression(AFnExpression node)
    {
	ATypedId typedId = (ATypedId)node.getTypedId();
	typedId.apply(this);
	Type argType = (Type)getOut(typedId.getTy());
	String argName = typedId.getId().getText();
	env.openScope();
	env.put(argName,new EnvEntry(argType));
	env.checkPoint();
	node.getExpression().apply(this);
	if (env.removed()) {
	    errorTable.add(node.getFn().getLine(),
			   new LinVarInClosureError());
	};
	Type bodyType = (Type)getOut(node.getExpression());
	if (bodyType != null) {
	    setOut(node,new FunType(argType,bodyType));
	};
	if (env.closeScope()) {
	    errorTable.add(node.getFn().getLine(),new LinVarsUnusedError());
	};
	position.put(node,node.getFn().getLine());
    }

    public void caseALfnExpression(ALfnExpression node)
    {
	ATypedId typedId = (ATypedId)node.getTypedId();
	typedId.apply(this);
	Type argType = (Type)getOut(typedId.getTy());
	String argName = typedId.getId().getText();
	env.openScope();
	env.put(argName,new EnvEntry(argType));
	node.getExpression().apply(this);
	Type bodyType = (Type)getOut(node.getExpression());
	if (bodyType != null) {
	    setOut(node,new LinFunType(argType,bodyType));
	};
	if (env.closeScope()) {
	    errorTable.add(node.getLfn().getLine(),new LinVarsUnusedError());
	};
	position.put(node,node.getLfn().getLine());
    }

    public void outAAppExpression(AAppExpression node)
    {
	Type fnType = (Type)getOut(node.getFn());
	Type actualArgType = (Type)getOut(node.getArg());

	if (fnType != null) {
	    if (fnType instanceof FunType) {
		Type formalArgType = ((FunType)fnType).getArg();
		if (actualArgType != null) {
		    if (!formalArgType.equal(actualArgType)) {
			errorTable.add(position.get(node.getFn()),
				       new WrongArgTypeError
				       (formalArgType.name(),
					actualArgType.name()));
		    }
		    else {
			setOut(node, ((FunType)fnType).getRes());
		    }
		}
	    }
	    else if (fnType instanceof LinFunType) {
		Type formalArgType = ((LinFunType)fnType).getArg();
		if (actualArgType != null) {
		    if (!formalArgType.equal(actualArgType)) {
			errorTable.add(position.get(node.getFn()),
				       new WrongArgTypeError
				       (formalArgType.name(),
					actualArgType.name()));
		    }
		    else {
			setOut(node, ((LinFunType)fnType).getRes());
		    }
		}
	    }
	    else {
		errorTable.add(position.get(node.getFn()),
			       new AppNotFunError(fnType.name()));
	    }
	};
	position.put(node,position.get(node.getFn()));
    }

    public void outAForkExpression(AForkExpression node)
    {
	Type childType = (Type)getOut(node.getChild());
	if (!childType.unlimited()) {
	    errorTable.add(node.getFork().getLine(),
			   new ForkLinearError(childType.name()));
	};
	setOut(node,(Type)getOut(node.getParent()));
	position.put(node,node.getFork().getLine());
    }

    public void outARequestExpression(ARequestExpression node)
    {
	Type portType = (Type)getOut(node.getTerm());
	if (portType != null) {
	    if (!(portType instanceof PortType)) {
		errorTable.add(node.getRequest().getLine(),
			       new RequestNotPortError(portType.name()));
	    }
	    else {
		setOut(node,((PortType)portType).getTy().dual());
	    }
	};
	position.put(node,node.getRequest().getLine());
    }

    public void outAAcceptExpression(AAcceptExpression node)
    {
	Type portType = (Type)getOut(node.getTerm());
	if (portType != null) {
	    if (!(portType instanceof PortType)) {
		errorTable.add(node.getAccept().getLine(),
			       new AcceptNotPortError(portType.name()));
	    }
	    else {
		setOut(node,((PortType)portType).getTy());
	    }
	};
	position.put(node,node.getAccept().getLine());
    }

    public void outASendExpression(ASendExpression node)
    {
	Type dataType = (Type)getOut(node.getData());
	Type chanType = (Type)getOut(node.getChan());
	if (chanType != null) {
	    chanType = chanType.unfold(tenv);
	    if (!(chanType instanceof OutputType)) {
		errorTable.add(node.getSend().getLine(),
			       new SendNotOutputError(chanType.name()));
	    }
	    else {
		Type chDataType = ((OutputType)chanType).getData();
		Type contType = ((OutputType)chanType).getCont();
		if (!chDataType.equal(dataType)) {
		    errorTable.add(node.getSend().getLine(),
			       new SendMismatchError(dataType.name(),
						     chanType.name()));
		};
		setOut(node,contType);
	    }
	};
	position.put(node,node.getSend().getLine());
    }

    public void outAReceiveExpression(AReceiveExpression node)
    {
	Type chanType = (Type)getOut(node.getTerm());
	if (chanType != null) {
	    chanType = chanType.unfold(tenv);
	    if (!(chanType instanceof InputType)) {
		errorTable.add(node.getReceive().getLine(),
			       new ReceiveNotInputError(chanType.name()));
	    }
	    else {
		Type chDataType = ((InputType)chanType).getData();
		Type contType = ((InputType)chanType).getCont();
		setOut(node,new TensorType(chDataType,contType));
	    }
	};
	position.put(node,node.getReceive().getLine());
    }

    public void outASelectExpression(ASelectExpression node)
    {
	String label = node.getId().getText();
	Type chanType = (Type)getOut(node.getTerm());
	if (chanType != null) {
	    chanType = chanType.unfold(tenv);
	    if (!(chanType instanceof ChoiceType)) {
		errorTable.add(node.getSelect().getLine(),
			       new SelectNotChoiceError(chanType.name()));
	    }
	    else {
		Type cType = ((ChoiceType)chanType).getBranches().get(label);
		if (cType == null) {
		    errorTable.add(node.getSelect().getLine(),
				   new SelectWrongLabelError(label,chanType.name()));
		}
		else {
		    setOut(node,cType);
		}
	    }
	};
	position.put(node,node.getSelect().getLine());
    }

    public void caseACaseExpression(ACaseExpression node)
    {
	node.getTerm().apply(this);
	Env envCopy = env.copy();
	ABranches branches = (ABranches)node.getBranches();
	ABranch first = (ABranch)branches.getBranch();
	Object[] rest = branches.getBarBranch().toArray();
	int numBranches = rest.length+1;
	String[] labels = new String[numBranches]; // from "case"
	String[] vars = new String[numBranches];   // from "case"
	Node[] exps = new Node[numBranches];       // from "case"
	Type[] contTypes = new Type[numBranches];  // from channel type
	Type[] expTypes = new Type[numBranches];   // from typechecking
	Type chanType = (Type)getOut(node.getTerm());
	if (chanType != null) {
	    chanType = chanType.unfold(tenv);
	    if (!(chanType instanceof BranchType)) {
		errorTable.add(node.getCase().getLine(),
			       new CaseNotBranchError(chanType.name()));
	    }
	    else {
		Hashtable<String,Type> brTypes = 
		    ((BranchType)chanType).getBranches();
		if (brTypes.size() != numBranches) {
		    errorTable.add(node.getCase().getLine(),
				   new CaseUnmatchedBranchesError(chanType.name()));
		};
		labels[0] = first.getLabel().getText();
		vars[0] = first.getVar().getText();
		exps[0] = first.getExpression();
		for (int i = 1; i < numBranches; i++) {
		    labels[i] = ((ABranch)(((ABarBranch)rest[i-1]).getBranch())).getLabel().getText();
		    vars[i] = ((ABranch)(((ABarBranch)rest[i-1]).getBranch())).getVar().getText();
		    exps[i] = ((ABranch)(((ABarBranch)rest[i-1]).getBranch())).getExpression();
		};
		boolean matchBrs = true;
		for (int i = 0; i < numBranches; i++) {
		    contTypes[i] = brTypes.get(labels[i]);
		    if (contTypes[i] == null) {
			errorTable.add(node.getCase().getLine(),
				   new CaseUnmatchedBranchesError(chanType.name()));
			matchBrs = false;
		    }
		};
		if (matchBrs) { // now check the branches
		    env.openScope();
		    env.put(vars[0],new EnvEntry(contTypes[0]));
		    exps[0].apply(this);
		    expTypes[0] = (Type)getOut(exps[0]);
		    env.closeScope();
		    Env envAfterFirstBranch = env.copy();
		    for (int i = 1; i < numBranches; i++) {
			env = envCopy;
			env.openScope();
			env.put(vars[i],new EnvEntry(contTypes[i]));
			exps[i].apply(this);
			expTypes[i] = (Type)getOut(exps[i]);
			env.closeScope();
			if (!env.equal(envAfterFirstBranch)) {
			    errorTable.add(node.getCase().getLine(),
					   new CaseEnvError());
			};
		    };
		    for (int i = 1; i < numBranches; i++) {
			if (!expTypes[i].equal(expTypes[0])) {
			    errorTable.add(node.getCase().getLine(),
					   new CaseBranchTypesError());
			}
		    };
		};
		setOut(node,expTypes[0]);
	    }
	};
	position.put(node,node.getCase().getLine());
    }

    public void outACloseExpression(ACloseExpression node)
    {
	Type chanType = (Type)getOut(node.getTerm());
	if (chanType != null) {
	    chanType = chanType.unfold(tenv);
	    if (!(chanType instanceof EndType)) {
		errorTable.add(node.getClose().getLine(),
			       new CloseNotEndError(chanType.name()));
	    }
	};
	setOut(node,new UnitType());
	position.put(node,node.getClose().getLine());
    }

    public void caseALetExpression(ALetExpression node)
    {
	env.openScope();
	node.getBinding().apply(this);
	node.getExpression().apply(this);
	setOut(node, getOut(node.getExpression()));
	if (env.closeScope()) {
	    errorTable.add(node.getLet().getLine(),new LinVarsUnusedError());
	};
	position.put(node,node.getLet().getLine());
    }

    public void outAIntLitTerm(AIntLitTerm node)
    {
	setOut(node, new IntType());
	position.put(node,node.getInt().getLine());
    }

    public void outABoolLitTerm(ABoolLitTerm node)
    {
	setOut(node, new BoolType());
	position.put(node,position.get(node.getBool()));
    }

    public void outAIdTerm(AIdTerm node) {
	EnvEntry entry = env.get(node.getId().getText());
	if (entry == null) {
	    errorTable.add(node.getId().getLine(),
			   new UndefinedIdError(node.getId().getText()));
	}
	else if (!(entry.available())) {
	    errorTable.add(node.getId().getLine(),
			   new UnavailableIdError(node.getId().getText()));
	} else {
	    setOut(node, entry.getType());
	};
	position.put(node,node.getId().getLine());
    }

    public void outAExpTerm(AExpTerm node)
    {
	setOut(node, getOut(node.getExpression()));
	position.put(node,position.get(node.getExpression()));
    }

    public void outAValBinding(AValBinding node)
    {
	String id = node.getId().getText();
	Type valTy = (Type)getOut(node.getExpression());
	if (valTy != null) {
	    if (env.put(id,new EnvEntry(valTy)) != null) {
		errorTable.add(node.getId().getLine(),
			       new ValRedefinedError(node.getId().getText()));
	    }
	};
	position.put(node,node.getEquals().getLine());
    }

    public void outASplitBinding(ASplitBinding node)
    {
	String left = node.getLeft().getText();
	String right = node.getRight().getText();
	Type expTy = (Type)getOut(node.getExpression());
	if (expTy != null) {
	    if (!(expTy instanceof TensorType)) {
		errorTable.add(node.getEquals().getLine(),
			       new SplitNotTensorError(expTy.name()));
	    }
	    else {
		Type leftTy = ((TensorType)expTy).getLeft();
		Type rightTy = ((TensorType)expTy).getRight();
		env.put(left,new EnvEntry(leftTy));
		env.put(right,new EnvEntry(rightTy));
	    }
	};
	position.put(node,node.getEquals().getLine());
    }

    public void caseAFunBinding(AFunBinding node)
    {
	ATypedId ta = (ATypedId)node.getTypedId();
	ta.getTy().apply(this);
	node.getTy().apply(this);
	Type argTy = (Type)getOut(ta.getTy());
	Type resTy = (Type)getOut(node.getTy());
	if (env.put(node.getId().getText(),new EnvEntry(new FunType(argTy,resTy))) != null) {
	    errorTable.add(node.getId().getLine(),
			   new ValRedefinedError(node.getId().getText()));
	}
	else {
	    env.openScope();
	    ta.apply(this);
	    env.checkPoint();
	    node.getExpression().apply(this);
	    if (env.removed()) {
		errorTable.add(node.getFun().getLine(),
			       new LinVarInClosureError());
	    };
	    Type bodyTy = (Type)getOut(node.getExpression());
	    if (bodyTy != null) {
		if (!bodyTy.equal(resTy)) {
		    errorTable.add(node.getId().getLine(),
			    new WrongResultTypeError(node.getId().getText(),
							    resTy.name(),
							    bodyTy.name()));
		}
	    };
	    if (env.closeScope()) {
		errorTable.add(node.getFun().getLine(),
			       new LinVarsUnusedError());
	    };
	};
	position.put(node,node.getFun().getLine());
    }

    public void outATrueBool(ATrueBool node) {
	position.put(node,node.getTrue().getLine());
    }

    public void outAFalseBool(AFalseBool node) {
	position.put(node,node.getFalse().getLine());
    }

    public void outABaseTy(ABaseTy node)
    {
	setOut(node, getOut(node.getBaseTy()));
	position.put(node,position.get(node.getBaseTy()));
    }

    public void outASessionTy(ASessionTy node)
    {
	setOut(node,(Type)getOut(node.getSessionTy()));
    }

    public void outAFunTy(AFunTy node)
    {
	setOut(node, new FunType((Type)getOut(node.getArg()),
				 (Type)getOut(node.getResult())));
	position.put(node,node.getFunArrow().getLine());
    }

    public void outALfunTy(ALfunTy node)
    {
	setOut(node, new LinFunType((Type)getOut(node.getArg()),
				 (Type)getOut(node.getResult())));
	position.put(node,node.getLinFunArrow().getLine());
    }

    public void outATensorTy(ATensorTy node)
    {
	setOut(node, new TensorType((Type)getOut(node.getLeft()),
				 (Type)getOut(node.getRight())));
	position.put(node,node.getTensor().getLine());
    }

    public void outAPortTy(APortTy node)
    {
	setOut(node, new PortType((Type)getOut(node.getTy())));
	position.put(node,node.getLbracket().getLine());
    }

    public void outAIntBaseTy(AIntBaseTy node) {
	setOut(node, new IntType());
	position.put(node,node.getIntTy().getLine());
    }

    public void outABoolBaseTy(ABoolBaseTy node) {
	setOut(node, new BoolType());
	position.put(node,node.getBoolTy().getLine());
    }

    public void outAUnitBaseTy(AUnitBaseTy node) {
	setOut(node, new UnitType());
	position.put(node,node.getUnitTy().getLine());
    }

    public void outACompoundBaseTy(ACompoundBaseTy node)
    {
	setOut(node,getOut(node.getTy()));
	position.put(node,position.get(node.getTy()));
    }

    public void outANamedSessionTy(ANamedSessionTy node)
    {
	if (!typeNames.contains(node.getId().getText())) {
	    errorTable.add(node.getId().getLine(),
			   new UndefinedTypeError(node.getId().getText()));
	};
	setOut(node,new NamedType(node.getId().getText(),false));
	position.put(node,node.getId().getLine());
    }

    public void outADualSessionTy(ADualSessionTy node)
    {
	if (!typeNames.contains(node.getId().getText())) {
	    errorTable.add(node.getId().getLine(),
			   new UndefinedTypeError(node.getId().getText()));
	};
	setOut(node,new NamedType(node.getId().getText(),true));
	position.put(node,node.getId().getLine());
    }

    public void outAEndSessionTy(AEndSessionTy node)
    {
	setOut(node,new EndType());
	position.put(node,node.getEndTy().getLine());
    }

    public void outAOutputSessionTy(AOutputSessionTy node)
    {
        setOut(node,new OutputType((Type)getOut(node.getBaseTy()),
				   (Type)getOut(node.getSessionTy())));
	position.put(node,node.getDot().getLine());
    }

    public void outAInputSessionTy(AInputSessionTy node)
    {
        setOut(node,new InputType((Type)getOut(node.getBaseTy()),
				  (Type)getOut(node.getSessionTy())));
	position.put(node,node.getDot().getLine());
    }

    public void outAChoiceSessionTy(AChoiceSessionTy node)
    {
	Hashtable<String,Type> branches = new Hashtable<String,Type>();
	ALabTys labTys = (ALabTys)node.getLabTys();
	ALabTy first = (ALabTy)labTys.getLabTy();
	branches.put(first.getId().getText(),
		     (Type)getOut(first.getSessionTy()));
	Object[] rest = labTys.getCommaLabTy().toArray();
	for (int i = 0; i < rest.length; i++) {
	    ACommaLabTy clt = (ACommaLabTy)rest[i];
	    ALabTy lt = (ALabTy)clt.getLabTy();
	    if (branches.put(lt.getId().getText(),
			     (Type)getOut(lt.getSessionTy())) != null) {
		errorTable.add(node.getPlus().getLine(),
			       new LabelReusedError(lt.getId().getText()));
	    };
	};
	setOut(node,new ChoiceType(branches));
	position.put(node,node.getPlus().getLine());
    }

    public void outABranchSessionTy(ABranchSessionTy node)
    {
	Hashtable<String,Type> branches = new Hashtable<String,Type>();
	ALabTys labTys = (ALabTys)node.getLabTys();
	ALabTy first = (ALabTy)labTys.getLabTy();
	branches.put(first.getId().getText(),
		     (Type)getOut(first.getSessionTy()));
	Object[] rest = labTys.getCommaLabTy().toArray();
	for (int i = 0; i < rest.length; i++) {
	    ACommaLabTy clt = (ACommaLabTy)rest[i];
	    ALabTy lt = (ALabTy)clt.getLabTy();
	    if (branches.put(lt.getId().getText(),
			     (Type)getOut(lt.getSessionTy())) != null) {
		errorTable.add(node.getAmp().getLine(),
			       new LabelReusedError(lt.getId().getText()));
	    };
	};
	setOut(node,new BranchType(branches));
	position.put(node,node.getAmp().getLine());
    }

    public void outATypedef(ATypedef node)
    {
	tenv.put(node.getId().getText(),
		 new TypeEnvEntry((Type)getOut(node.getTy())));
    }

    public void outATypedId(ATypedId node)
    {
        if (env.put(node.getId().getText(),
		    new EnvEntry((Type)getOut(node.getTy()))) != null) {
            errorTable.add(node.getId().getLine(),
                           new ValRedefinedError(node.getId().getText()));
        }
    }

}
