/****************************************************************************/
/*                                                                          */
/*                   Copyright 2006 University of Glasgow                   */
/*                                                                          */
/*                           All rights reserved.                           */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/* FILE          : Env.java                                                 */
/* DESCRIPTION   : Environment of identifier bindings                       */
/*                                                                          */
/* AUTHOR        : S.J.Gay                                                  */
/* DATE          : 21st December 2005                                       */
/*                                                                          */
/* LAST MODIFIED : S.J.Gay                                                  */
/* DATE          : 4th January 2006                                         */
/****************************************************************************/

package fls.env;

import fls.types.*;
import java.util.*;

/* An environment is a vector of hash tables, used as a stack.
The 'put' method adds an entry to the hash table at the top of the stack
(first element of the vector). The 'get' method searches down the stack.
The 'openScope' method creates a new hash table on the top of the stack.
The 'closeScope' method removes the top-most hash table. 

During typechecking, 'get' is used to discover the type of an identifier
when it is used in an expression. So if the type is linear, we mark 
the identifier as unavailable.

The 'checkPoint' method resets a flag which is used to indicate whether
any identifiers have been removed from anywhere other than the current scope.
The 'removed' method returns the value of this flag. These methods are used
when typechecking linear function definitions, to work out whether or
not any linear identifiers occur free in the body of the function.
*/

public class Env {

    Vector<Hashtable<String,EnvEntry>> stack;

    boolean rem = false;

    public Env() {
	stack = new Vector<Hashtable<String,EnvEntry>>(1,1);
	stack.addElement(new Hashtable<String,EnvEntry>());
    }

    public Env(Vector<Hashtable<String,EnvEntry>> st) {
	stack = st;
    }

    public Vector<Hashtable<String,EnvEntry>> getStack() {
	return stack;
    }

    public void openScope() {
	stack.addElement(new Hashtable<String,EnvEntry>());
    }

    // Returns true if there are unused linear variables in the current scope
    public boolean closeScope() {
	Hashtable<String,EnvEntry> env = stack.get(stack.size()-1);
	stack.removeElementAt(stack.size()-1);
	Enumeration<EnvEntry> envEnum = env.elements();
	while (envEnum.hasMoreElements()) {
	    EnvEntry entry = envEnum.nextElement();
	    if ((!entry.getType().unlimited()) && entry.available())
		return true;
	};
	return false;
    }

    public EnvEntry put(String n, EnvEntry e) {
	return stack.elementAt(stack.size()-1).put(n,e);
    }

    public EnvEntry get(String n) {
	for (int i = stack.size()-1; i >= 0; i--) {
	    EnvEntry e = stack.elementAt(i).get(n);
	    if (e != null) {
		EnvEntry r = new EnvEntry(e.getType(),e.available());
		if (e.available()) {
		    if (!(e.getType().unlimited())) {
			e.makeUnavailable();
			if (i < stack.size()-1) rem = true;
		    };
		};
		return r;
	    };
	};
	return null;
    }

    public void checkPoint() {
	rem = false;
    }

    public boolean removed() {
	return rem;
    }

    public boolean linearRemaining() {
	Enumeration<Hashtable<String,EnvEntry>> stackEnum = stack.elements();
	while (stackEnum.hasMoreElements()) {
	    Hashtable<String,EnvEntry> env = stackEnum.nextElement();
	    Enumeration<EnvEntry> envEnum = env.elements();
	    while (envEnum.hasMoreElements()) {
		EnvEntry entry = envEnum.nextElement();
		if ((!entry.getType().unlimited()) && entry.available())
		    return true;
	    }
	};
	return false;
    }

    public Env copy() {
	Vector<Hashtable<String,EnvEntry>> st = 
	    new Vector<Hashtable<String,EnvEntry>>(1,1);
	for (int i = 0; i < stack.size(); i++) {
	    Hashtable<String,EnvEntry> env = stack.elementAt(i);
	    Enumeration<String> keys = env.keys();
	    Hashtable<String,EnvEntry> envCopy = 
		new Hashtable<String,EnvEntry>();
	    while (keys.hasMoreElements()) {
		String k = keys.nextElement();
		EnvEntry v = env.get(k);
		EnvEntry vCopy = new EnvEntry(v.getType(),v.available());
		envCopy.put(k,vCopy);
	    };
	    st.addElement(envCopy);
	};
	return new Env(st);
    }

    public boolean equal(Env other) {
	Vector<Hashtable<String,EnvEntry>> otherStack = other.getStack();
	if (otherStack.size() != stack.size()) return false;
	for (int i = 0; i < stack.size(); i++) {
	    Hashtable<String,EnvEntry> env = stack.elementAt(i);
	    Hashtable<String,EnvEntry> otherEnv = otherStack.elementAt(i);
	    if (env.size() != otherEnv.size()) return false;
	    Enumeration<String> keys = env.keys();
	    while (keys.hasMoreElements()) {
		String k = keys.nextElement();
		EnvEntry entry = env.get(k);
		EnvEntry otherEntry = otherEnv.get(k);
		if (otherEntry == null) return false;
		if (!entry.getType().equal(otherEntry.getType())) return false;
		if (entry.available() != otherEntry.available()) return false;
	    }
	};
	return true;
    }

}
